--/****************************************************************************
--Software: MIDI XML CLASS
--Version:  1.1
--Date:     2005/12/31
--Author:   Valentin Schmidt
--Contact:  fluxus@freenet.de
--License:  Freeware
--
-- Requirements/Dependencies:
-- o MIDI CLASS (parent script "MIDI")
-- o xtra "xmlparser" (of course could be rewritten to use other parser like flash or a lingo parser script instead)
--
--****************************************************************************/


property pTracks          --array of tracks, where each track is array of message strings
property pTimebase        --timebase = ticks per frame (quarter note)
property pTempo           --tempo as integer (0 for unknown)
property pTempoMsgNum     --position of tempo event in track 0

property pLF
property pCR

property ancestor

--/****************************************************************************
--*                                                                           *
--*                              Public methods                               *
--*                                                                           *
--****************************************************************************/

-----------------------------------------------------------------
-- creates (or resets to) new empty MIDI song
-----------------------------------------------------------------
on new (me,timebase) --open
  
  if voidP(timebase) then timebase=480
  me.ancestor=new(script "MIDI",timebase)
  
  me.pTempo = 0 --125000 = 120 bpm
  me.pTimebase = timebase
  me.pTracks = []
  
  pLF=chr(10) --\n
  pCR=chr(13) --\r
  
  return me
end

-----------------------------------------------------------------
-- returns MIDI XML representation (v0.8, http:--www.musicxml.org/dtds/midixml.dtd)
-----------------------------------------------------------------
on getXml (me,ttype,ret) --0:absolute, 1:delta
  if voidP(ttype) then ttype=0
  if voidP(ret) then ret=pLF
  
  ch=VOID
  p=VOID
  v=VOID
  n=VOID
  c=VOID
  
  q=QUOTE
  
  
  tracks = me.pTracks
  tc = count(tracks)
  type = integer(tc>1)
  timebase = me.pTimebase
  
  stamp=["Delta","Absolute"][2-(ttype=1)]
  
  xml = "<?xml version="&q&"1.0"&q&" encoding="&q&"ISO-8859-1"&q&"?>"&ret&\
"<!DOCTYPE MIDIFile PUBLIC"&ret&\
q&"-//Recordare//DTD MusicXML 0.9 MIDI//EN"&q&ret&\
q&"http://www.musicxml.org/dtds/midixml.dtd"&q&">"&ret&\
"<MIDIFile>"&ret&\
"<Format>"&type&"</Format>"&ret&\
"<TrackCount>"&tc&"</TrackCount>"&ret&\
"<TicksPerBeat>"&timebase&"</TicksPerBeat>"&ret&\
"<TimestampType>"&stamp&"</TimestampType>"&ret
  
  repeat with i = 1 to tc
    xml=xml& "<Track Number="&q&i&q&">"&ret
    track = pTracks[i]
    mc = count(track)
    last = 0
    
    repeat with j = 1 to mc
      msg = explode(" ",track[j])
      t = integer(msg[1])
      if (ttype=1) then--delta
        dt = t - last
        last = t
      end if
      xml=xml& "  <Event>"&ret
      if ttype=1 then xml=xml& "    <Delta>"&dt&"</Delta>"&ret
      else xml=xml&            "    <Absolute>"&t&"</Absolute>"&ret
      xml=xml& "    "
      
      case (msg[2]) of
        "PrCh":
          do(msg[3]) -- ch
          do(msg[4]) -- p
          xml=xml& "<ProgramChange Channel="&q&ch&q&" Number="&q&p&q&"/>"&ret
          
        "On","Off":
          do(msg[3]) -- ch
          do(msg[4]) -- n
          do(msg[5]) -- v
          xml=xml& "<Note"&msg[2]&" Channel="&q&ch&q&" Note="&q&n&q&" Velocity="&q&v&q&"/>"&ret
          
        "PoPr":
          do(msg[3]) -- ch
          do(msg[4]) -- n
          do(msg[5]) -- v
          xml=xml& "<PolyKeyPressure Channel="&q&ch&q&" Note="&q&n&q&" Pressure="&q&v&q&"/>"&ret
          
        "Par":
          do(msg[3]) -- ch
          do(msg[4]) -- c
          do(msg[5]) -- v
          xml=xml& "<ControlChange Channel="&q&ch&q&" Control="&q&c&q&" Value="&q&v&q&"/>"&ret
          
        "ChPr":
          do(msg[3]) -- ch
          do(msg[4]) -- v
          xml=xml& "<ChannelKeyPressure Channel="&q&ch&q&" Pressure="&q&v&q&"/>"&ret
          
        "Pb":
          do(msg[3]) -- ch
          do(msg[4]) -- v
          xml=xml& "<PitchBendChange Channel="&q&ch&q&" Value="&q&v&q&"/>"&ret
          
        "Seqnr":
          xml=xml& "<SequenceNumber Value="&q&msg[3]&q&"/>"&ret
          
        "Meta":
          txttypes = ["Text","Copyright","TrkName","InstrName","Lyric","Marker","Cue"]
          mtype = msg[3]
          
          pos = txttypes.getPos(mtype)
          if (pos>0) then
            tags = ["TextEvent","CopyrightNotice","TrackName","InstrumentName","Lyric","Marker","CuePoint"]
            tag = tags[pos]
            lin = track[j]
            start = offset(q,lin)+1
            ende = start+offset(q,lin.char[start+1..length(lin)])-1
            txt = lin.char[start..ende]
            xml=xml& "<"&tag&">"&txt&"</"&tag&">"&ret --htmlspecialchars
          else if (mtype = "TrkEnd") then
            xml=xml& "<EndOfTrack/>"&ret
          else if (mtype = "0x20" OR mtype = "0x21") then -- ChannelPrefix ???
            xml=xml& "<MIDIChannelPrefix Value="&q&msg[4]&q&"/>"&ret
          end if
          
        "Tempo":
          xml=xml& "<SetTempo Value="&q&msg[3]&q&"/>"&ret
          
        "SMPTE":
          xml=xml& "<SMPTEOffset TimeCodeType="&q&1&q&" Hour="&q&msg[3]&q&" Minute="&q&msg[4]&q&" Second="&q&msg[5]&q&" Frame="&q&msg[6]&q&" FractionalFrame="&q&msg[7]&q&"/>"&ret --TimeCodeType???
          
        "TimeSig":
          ts = explode("/",msg[3])
          xml=xml& "<TimeSignature Numerator="&q&ts[1]&q&" LogDenominator="&q&integer(log(integer(ts[2]))/log(2))&q&" MIDIClocksPerMetronomeClick="&q&msg[4]&q&" ThirtySecondsPer24Clocks="&q&msg[5]&q&"/>"&ret
          
        "KeySig":
          mode = integer(msg[4]<>"major")
          xml=xml& "<KeySignature Fifths="&q&msg[3]&q&" Mode="&q&mode&q&"/>"&ret -- ???
          
        "SeqSpec":
          lin = track[j]
          start = offset("SeqSpec",lin)+8
          data = lin.char[start..length(lin)]
          xml=xml& "<SequencerSpecific>"&data&"</SequencerSpecific>"&ret
          
        "SysEx":
          str = ""
          --for (k=3k<count(msg)k++) str .= msg[k]." "
          repeat with k=4 to count(msg)
            str=str& msg[k] & " "
          end repeat
          str = trim(strtoupper(str))
          xml=xml& "<SystemExclusive>"&str&"</SystemExclusive>"&ret
          
          --                  TODO:
          --                  <AllSoundOff Channel="9"/>
          --                  <ResetAllControllers Channel="9"/>
          --                  <LocalControl Channel="9" Value="on"/>
          --                  <AllNotesOff Channel="9"/>
          --                  <OmniOff Channel="9"/>
          --                  <OmniOn Channel="9"/>
          --                  <MonoMode Channel="9" Value="5"/>
          --                  <PolyMode Channel="9"/>
          
        otherwise:
          me._err(">>> unknown event: ".msg[2])
          halt()
      end case
      
      xml=xml& "  </Event>"&ret
    end repeat
    xml=xml& "</Track>"&ret
  end repeat
  xml=xml& "</MIDIFile>"
  return xml
end

-----------------------------------------------------------------
-- import MIDI XML representation
-- (so far only a subset of http://www.musicxml.org/dtds/midixml.dtd (v0.9), see documentation)
-----------------------------------------------------------------
on importXml (me,xmlStr)
  if NOT (xmlStr starts "<?xml") then me._err(">>> no valid MIDI XML!")
  parserObject = xtra("xmlparser").new()
  --parserObj.ignorewhitespace(FALSE) --???
  errorCode = parserObject.parseString(xmlStr)
  
  mf = parserObject.child[1]
  repeat with i = 1 to mf.count(#child)
    ch=mf.child[i]
    nam=strtoupper(ch.name)
    case (nam) of
        --"MIDIFILE":
        
      "FORMAT":
        --me.pType = integer(ch.child[1].text) -- _chData ???
        
      "TRACKCOUNT":
        --
        
      "TICKSPERBEAT":
        me.pTimebase = integer(ch.child[1].text)
        
      "TIMESTAMPTYPE":
        ttype = ch.child[1].text--DELTA or ABSOLUTE
        
      "TRACK":
        me.newTrack()
        if (ttype="Delta") then tLast = 0
        
        msgCnt=ch.count(#child)
        repeat with j= 1 to msgCnt
          evt=ch.child[j]
          if evt.name="Event" then
            
            time = integer(evt.child[1].child[1].text)
            if (ttype="Delta") then
              time=time+ tLast
              tLast=time
            end if
            a=evt.child[2]
            typ=strtoupper(a.name)      
            
            case (typ) of
              "PROGRAMCHANGE":
                c=a.attributeValue["Channel"]
                p=a.attributeValue["Number"]
                msg = time && "PrCh ch="&c&" p="&p
              "NOTEON":
                c=a.attributeValue["Channel"]
                n=a.attributeValue["Note"]
                vel=a.attributeValue["Velocity"]
                msg = time && "On ch="&c&" n="&n&" v="&vel
              "NOTEOFF":
                c=a.attributeValue["Channel"]
                n=a.attributeValue["Note"]
                vel=a.attributeValue["Velocity"]
                msg = time && "Off ch="&c&" n="&n&" v="&vel
              "POLYKEYPRESSURE":
                c=a.attributeValue["Channel"]
                n=a.attributeValue["Note"]
                p=a.attributeValue["Pressure"]
                msg = time && "PoPr ch="&c&" n="&n&" v="&p
              "CONTROLCHANGE":
                c=a.attributeValue["Channel"]
                ct=a.attributeValue["Control"]
                v=a.attributeValue["Value"]
                msg = time && "Par ch="&c&" c="&ct&" v="&v
              "CHANNELKEYPRESSURE":
                c=a.attributeValue["Channel"]
                p=a.attributeValue["Pressure"]
                msg = time && "PoPr ch="&c&" v="&p
              "PITCHBENDCHANGE":
                c=a.attributeValue["Channel"]
                v=a.attributeValue["Value"]
                msg = time && "Pb ch="&c&" v="&v
              "SEQUENCENUMBER":
                v=a.attributeValue["Value"]
                msg = time && "Seqnr "&v
              "TEXTEVENT","COPYRIGHTNOTICE","COPYRIGHTNOTICE","TRACKNAME","INSTRUMENTNAME","LYRIC","MARKER","CUEPOINT":
                tags = ["TEXTEVENT","COPYRIGHTNOTICE","TRACKNAME","INSTRUMENTNAME","LYRIC","MARKER","CUEPOINT"]
                txttypes = ["Text","Copyright","TrkName","InstrName","Lyric","Marker","Cue"]
                type = txttypes[tags.getPos(typ)]
                dat=a.child[1].text
                msg = time && "Meta" && type && QUOTE&dat&QUOTE --???
              "ENDOFTRACK":
                msg = time && "Meta TrkEnd"
              "MIDICHANNELPREFIX":
                dat=a.child[1].text
                if (integer(dat)<10) then dat="0"&dat
                msg = time && "Meta 0x20" && dat
              "SETTEMPO":
                tempo = integer(a.attributeValue["Value"])
                msg = time && "Tempo" && tempo
                tn=count(me.pTracks)
                if (tn=0 AND time=0) then
                  me.pTempo = tempo
                  me.pTempoMsgNum = j --count(this->tracks[tn])--???
                end if
              "SMPTEOFFSET":
                h=a.attributeValue["Hour"]
                m=a.attributeValue["Minute"]
                s=a.attributeValue["Second"]
                f=a.attributeValue["Frame"]
                ff=a.attributeValue["FractionalFrame"]                
                msg = time && "SMPTE" && h && m && s && f && ff
              "TIMESIGNATURE":
                z=integer(a.attributeValue["Numerator"])
                t=integer(power(2,integer(a.attributeValue["LogDenominator"])))
                mi=a.attributeValue["MIDIClocksPerMetronomeClick"]
                th=a.attributeValue["ThirtySecondsPer24Clocks"]
                msg = time  && "TimeSig "&z&"/"&t && mi && th
              "KEYSIGNATURE":
                f=a.attributeValue["Fifths"]
                g=["major","minor"][2-(a.attributeValue["Mode"]=0)]
                msg = time && "KeySig" && f && g
              "SEQUENCERSPECIFIC":
                dat=a.child[1].text
                msg = time && "SeqSpec" && dat
              "SYSTEMEXCLUSIVE":
                dat = strtolower(a.child[1].text)
                msg = time && "SysEx f0 " & dat
                
                
              otherwise:
                next repeat --ignore unknown events
            end case
            
            me.addMsg(count(me.pTracks),msg)
            
          end if
        end repeat -- END EVENTS IN TRACK
        
        
      otherwise:
        nothing -- ignore unknown XML elements
    end case
    
  end repeat
  
end
